/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */
#include "CountWithSpectro.h"
#include "controllers/common/family/Families.h"
#include "PersistenceServices/DataFile/NexusDataFile.h"
#include <boost/date_time/posix_time/posix_time.hpp>
#include <common/base/Date.h>

namespace d22special {

const std::string CountWithSpectro::TYPE = "count_with_spectro";
const int32 CountWithSpectro::XSIZE = 1044; //should be retrieved directly from the driver...
const int32 CountWithSpectro::YSIZE = 1044;

CountWithSpectro::CountWithSpectro(const std::string & name) :
		ExperimentController(name), controller::Stoppable(this) {

	setFamily(family::SAMPLE_ENVIRONMENT);

	registerFunction(TYPE);

	spectro.init(this, "spectro");
	count.init(this, "count");
	spectroSwitcher.init(this, "spectroSwitcher");

	integrationTime.init(this, SAVE, "integration_time"); //in microseconds!
	numSpectroMeasures.init(this, SAVE, "num_spectro");
	timeType.init(this, SAVE, "time_type");
	countTime.init(this, SAVE, "count_time");

	xSize.init(this, SAVE, "x_size");
	ySize.init(this, SAVE, "y_size");
	zSize.init(this, SAVE, "z_size");

	subtitle.init(this, SAVE, "subtitle");
	useSwitcher.init(this, SAVE, "use_switcher");

	yDataArray.init(this, NOSAVE, "yData_array");
	initialTime.init(this, NOSAVE, "initial_time");
	measureDuration.init(this, NOSAVE, "measure_duration");

	yDataArraySecondChannel.init(this, NOSAVE, "yDataArray_secondChannel");
}

CountWithSpectro::~CountWithSpectro() {

}

void CountWithSpectro::postConfiguration() {
	xSize = XSIZE;
	ySize = numSpectroMeasures();
	zSize = 1;
	subtitle = count->subtitle();

	registerRefresher(numSpectroMeasures, &CountWithSpectro::refreshNumSpectroMeasures, this);
	registerUpdater(count->subtitle, &CountWithSpectro::updateSubtitle, this);
	registerProgression(count, &CountWithSpectro::updateProgression, this);

//	initSwitcherArrays(); //init both arrays to avoid errors
}

void CountWithSpectro::refreshNumSpectroMeasures(int32 number) {
	ySize = number; //This property is only used to save the data in nexus file

	//initialize here the dynamic properties
	measureDuration.resize(number);
}

void CountWithSpectro::updateProgression() {
	commandProgression = count->commandProgression();
}

void CountWithSpectro::updateSubtitle() {
	subtitle = count->subtitle();
}

void CountWithSpectro::start() {

	if (!useSwitcher()) {

		//Time already in microseconds!
		spectro->integrationTime = integrationTime();

		//check the total time of spectro measurement is smaller than the count time
		float64 spectroTime = integrationTime() * 0.000001 * numSpectroMeasures(); //should be in seconds
		float64 countTime = getCountTimeInSec(); //count->time.setpoint();

		if (spectroTime <= countTime) {

			setupCountParams();

			initArrays(); //before initArray()

			commandStatus.setRunning();
			spectro->mode = QE65000Controller::COUNT_SINGLE;

			//Launch count and spectro at the same time
			// Run the threads in parallel in a group
			boost::thread_group group;

			group.create_thread(boost::bind(&CountWithSpectro::startCount, this));
			group.create_thread(boost::bind(&CountWithSpectro::startSpectro, this));

			// Wait for the termination of the threads
			group.join_all();

			//Get data
			saveSpectroData();

			commandStatus.setIdle();
			spectro->mode = QE65000Controller::IDLE;
		}

	} else {
		//Suppose here we also keep the same rule as before (but now is double)
		spectroSwitcher->integrationTime = integrationTime();
		spectroSwitcher->numberOfCycles = 1;

		//check the total time of spectro measurement is smaller than the count time
		float64 spectroTime = integrationTime() * 0.000001 * numSpectroMeasures() * 2; //!!!!!!!!!!!!!!
		float64 countTime = getCountTimeInSec();

		if (spectroTime <= countTime) {

			setupCountParams();

			initArrays();

			commandStatus.setRunning();
			spectro->mode = QE65000Controller::COUNT_SWITCHER;

			//Launch count and spectro at the same time
			// Run the threads in parallel in a group
			boost::thread_group group;

			group.create_thread(boost::bind(&CountWithSpectro::startCount, this));
			group.create_thread(boost::bind(&CountWithSpectro::startSwitcher, this));

			// Wait for the termination of the threads
			group.join_all();

			//Get data
			saveSwitcherSpectroData();

			commandStatus.setIdle();
			spectro->mode = QE65000Controller::IDLE;
		}
	}
}

void CountWithSpectro::stop() {
	if (commandStatus.isRunning()) {
		count->stopParallel();
		spectro->stopParallel();
	}
	commandStatus.setWarning();

}

void CountWithSpectro::startCount() {
	count->startCommand();
}

void CountWithSpectro::startSpectro() {

	common::Date startTime = common::Date();
	initialTime = startTime.toString("%H:%M:%S");
	common::Date formerTime = startTime;

	for (int i = 0; i < numSpectroMeasures(); i++) {
		if (commandStatus.isRunning()) {

			common::Date time = common::Date();
			common::Duration duration = time - formerTime;
			measureDuration.set(i, duration.getMilliseconds());

			spectro->startCommand();

			int32 ySize = YSIZE;
			float64* tempYData = spectro->yData();
			for (int32 j = 0; j < ySize; j++) {
				m_yDataArray[ySize * i + j] = tempYData[j];
			}

			formerTime = time;
		}
	}
}

void CountWithSpectro::startSwitcher() {
	common::Date startTime = common::Date();
	initialTime = startTime.toString("%H:%M:%S");
	common::Date formerTime = startTime;

	for (int i = 0; i < numSpectroMeasures(); i++) {
		if (commandStatus.isRunning()) {

			common::Date time = common::Date();
			common::Duration duration = time - formerTime;
			measureDuration.set(i, duration.getMilliseconds());

			spectroSwitcher->startCommand();

			int32 ySize = YSIZE;
			float64* tempLeftYData = spectroSwitcher->yLeftData();
			for (int32 j = 0; j < ySize; j++) {
				m_yDataArray[ySize * i + j] = tempLeftYData[j];
			}

			float64* tempRightYData = spectroSwitcher->yRightData();
			for (int32 j = 0; j < ySize; j++) {
				m_yDataArraySecondChannel[ySize * i + j] = tempRightYData[j];
			}

			formerTime = time;
		}
	}
}

void CountWithSpectro::setupCountParams() {
	count->time.setpoint = countTime();
	count->timeType = timeType();
	count->subtitle = subtitle();
}

void CountWithSpectro::initArrays() {
	if (m_yDataArray) {
		delete[] m_yDataArray;
	} else if (m_yDataArraySecondChannel) {
		delete[] m_yDataArraySecondChannel;
	}

	int32 totalSize = YSIZE * numSpectroMeasures();
	m_yDataArray = new int32[totalSize];
	memset(m_yDataArray, 0, totalSize * sizeof(int32)); //set to zero
	yDataArray.update(m_yDataArray);
	yDataArray.setSize(totalSize);

	m_yDataArraySecondChannel = new int32[totalSize];
	memset(m_yDataArraySecondChannel, 0, totalSize * sizeof(int32)); //set to zero
	yDataArraySecondChannel.update(m_yDataArraySecondChannel);
	yDataArraySecondChannel.setSize(totalSize);

}

void CountWithSpectro::saveSpectroData() {
	int32 totalSize = YSIZE * numSpectroMeasures();
	yDataArray.update(m_yDataArray);
	yDataArray.setSize(totalSize);

	int32 numor = any_cast<int32>(count->getValue("numor"));
	NexusDataFile::appendTo(numor, "NxD22SpecialServerFile.xml", "NxD22SpecialClientFile.xml");
}

void CountWithSpectro::saveSwitcherSpectroData() {
	int32 totalSize = YSIZE * numSpectroMeasures();
	yDataArray.update(m_yDataArray);
	yDataArray.setSize(totalSize);

	yDataArraySecondChannel.update(m_yDataArraySecondChannel);
	yDataArraySecondChannel.setSize(totalSize);

	int32 numor = any_cast<int32>(count->getValue("numor"));
	NexusDataFile::appendTo(numor, "NxD22SpecialServerFile.xml", "NxD22SpecialClientFile.xml");

}

float64 CountWithSpectro::getCountTimeInSec() {
	float64 time;
	if (timeType() == "h") {
		time = countTime() * 3600.;
	} else if (timeType() == "m") {
		time = countTime() * 60.;
	} else if (timeType() == "s") {
		time = countTime();
	}

	return time;
}

}

