/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CAENPhaController.h"
#include "drivers/global/DriversCommands.h"
#include "controllers/common/family/Families.h"
#include <boost/thread/thread.hpp>

using namespace std;
using namespace boost;

namespace npp {

const string CAENPhaController::TYPE = "pha_controller";

/*
 * Constructor
 */
CAENPhaController::CAENPhaController(const string& name) : CAENAdcController(name) {

	setFamily(family::ACQUISITION, family::SETTING);

	nbBits.init(this, NOSAVE, "nb_bits");
	signalDecayTime.init(this, NOSAVE, "decay_time");
	trapezoidFlatTop.init(this, NOSAVE, "trapezoid_flat");
	trapezoidRiseTime.init(this, NOSAVE, "trapezoid_rise");
	flatTopDelay.init(this, NOSAVE, "flat_top");
	triggerSmoothing.init(this, NOSAVE, "trigger_smoothing");
	signalRiseTime.init(this, NOSAVE, "rise_time");
	triggerThreshold.init(this, NOSAVE, "trigger_threshold");
	baselineMean.init(this, NOSAVE, "base_mean");
	peakMean.init(this, NOSAVE, "peak_mean");
	peakHoldOff.init(this, NOSAVE, "peak_holdoff");
	baselineHoldOff.init(this, NOSAVE, "base_holdoff");
	triggerHoldOff.init(this, NOSAVE, "trigger_holdoff");
	triggerWindowSize.init(this, NOSAVE, "trigger_window_size");
	triggerWindowEnable.init(this, NOSAVE, "trigger_window_enable");
	digitalGain.init(this, NOSAVE, "digital_gain");
	trapezoidGain.init(this, NOSAVE, "trapezoid_gain");
	decimation.init(this, NOSAVE, "decimation");
	useFirstDerivative.init(this, NOSAVE, "use_first_derivative");

	m_dgtzDriver.init(this, "pha_driver");

	registerFunction(TYPE);
}

/*
 * Destructor
 */
CAENPhaController::~CAENPhaController() {

}

/*
 * postConfiguration
 */
void CAENPhaController::postConfiguration() {

	registerRefresher(recordLength, &CAENPhaController::refreshRecordLengthProperty, this);
	registerRefresher(nbChannels, &CAENPhaController::refreshNbChannelsProperty, this);
	registerRefresher(dppAcqMode, &CAENPhaController::refreshDppAcqModeProperty, this);
	registerRefresher(displayChannel, &CAENPhaController::refreshDisplayChannelProperty, this);
	registerRefresher(signalDecayTime, &CAENPhaController::refreshSignalDecayTime, this);
	registerRefresher(trapezoidFlatTop, &CAENPhaController::refreshTrapezoidFlatTopProperty, this);
	registerRefresher(trapezoidRiseTime, &CAENPhaController::refreshTrapezoidRiseTimeProperty, this);
	registerRefresher(flatTopDelay, &CAENPhaController::refreshFlatTopDelayProperty, this);
	registerRefresher(triggerSmoothing, &CAENPhaController::refreshTriggerSmoothing, this);
	registerRefresher(signalRiseTime, &CAENPhaController::refreshSignalRiseTimeProperty, this);
	registerRefresher(triggerThreshold, &CAENPhaController::refreshTriggerThresholdProperty, this);
	registerRefresher(baselineMean, &CAENPhaController::refreshBaselineMean, this);
	registerRefresher(peakMean, &CAENPhaController::refreshPeakMean, this);
	registerRefresher(peakHoldOff, &CAENPhaController::refreshPeaklineHoldOffProperty, this);
	registerRefresher(baselineHoldOff, &CAENPhaController::refreshBaselineHoldOffProperty, this);
	registerRefresher(triggerHoldOff, &CAENPhaController::refreshTriggerHoldOffProperty, this);
	registerRefresher(triggerWindowSize, &CAENPhaController::refreshTriggerWindowSizeProperty, this);
	registerRefresher(triggerWindowEnable, &CAENPhaController::refreshTriggerWindowEnableProperty, this);
	registerRefresher(digitalGain, &CAENPhaController::refreshDigitalGainProperty, this);
	registerRefresher(trapezoidGain, &CAENPhaController::refreshTrapezoidGainProperty, this);
	registerRefresher(decimation, &CAENPhaController::refreshDecimationProperty, this);
	registerRefresher(channelActive, &CAENPhaController::refreshChannelActiveProperty, this);
	registerRefresher(dcOffset, &CAENPhaController::refreshDcOffsetProperty, this);
	registerRefresher(preTrigger, &CAENPhaController::refreshPreTriggerProperty, this);
	registerRefresher(pulsePolarity, &CAENPhaController::refreshPulsePolarityProperty, this);
	registerRefresher(useFirstDerivative, &CAENPhaController::refreshUseFirstDerivativeProperty, this);

	updateBoardProperties();
}

void CAENPhaController::refreshRecordLengthProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->recordLength = aValue;
}

void CAENPhaController::refreshNbChannelsProperty(int32 aValue)  throw (CannotSetValue) {

	minRange = 0;
	maxRange = aValue - 1;
	m_dgtzDriver->nbChannels = aValue;
	if (displayChannel() >= aValue)
		displayChannel = aValue - 1;
}

void CAENPhaController::refreshDppAcqModeProperty(int32 aValue)  throw (CannotSetValue) {

	m_dgtzDriver->dppAcqMode = aValue;
	boost::thread td(boost::bind(&CAENPhaController::initBoard, this));
}

void CAENPhaController::refreshDisplayChannelProperty(int32 aChannel) throw (CannotSetValue) {

	float64 decayTime = m_dgtzDriver->signalDecayTime.get(aChannel) / 1000.0;
	if (decayTime == 0.0)
		decayTime = 20.0;
	signalDecayTime.update(decayTime);

	float64 flatTop = m_dgtzDriver->trapezoidFlatTop.get(aChannel) / 1000.0;
	if (flatTop == 0.0)
		flatTop = 1.0;
	trapezoidFlatTop.update(flatTop);

	float64 trapRise = m_dgtzDriver->trapezoidRiseTime.get(aChannel) / 1000.0;
	if (trapRise == 0.0)
		trapRise = 1.0;
	trapezoidRiseTime.update(trapRise);

	float64 peaking = m_dgtzDriver->flatTopDelay.get(aChannel) / 1000.0;
	if (peaking == 0.0)
		peaking = 0.5;
	flatTopDelay.update(peaking);

	int32 trigAvrg = m_dgtzDriver->triggerSmoothing.get(aChannel);
	if (trigAvrg == 0)
		trigAvrg = 2;
	triggerSmoothing.update(trigAvrg);

	float64 riseTime = m_dgtzDriver->signalRiseTime.get(aChannel) / 1000.0;
	if (riseTime == 0.0)
		riseTime = 0.2;
	signalRiseTime.update(riseTime);

	int32 trigTrsd = m_dgtzDriver->triggerThreshold.get(aChannel);
	if (trigTrsd == 0)
		trigTrsd = 200;
	triggerThreshold.update(trigTrsd);

	int32 baseAvrg = m_dgtzDriver->baselineMean.get(aChannel);
	if (baseAvrg == 0)
		baseAvrg = 2;
	baselineMean.update(baseAvrg);

	int32 peakAvrg = m_dgtzDriver->peakMean.get(aChannel);
	if (peakAvrg == 0)
		peakAvrg = 1;
	peakMean.update(peakAvrg);

	float64 peakHold = m_dgtzDriver->peakHoldOff.get(aChannel) / 1000.0;
	if (peakHold == 0.0)
		peakHold = 0.5;
	peakHoldOff.update(peakHold);

	float64 baseHold = m_dgtzDriver->baselineHoldOff.get(aChannel) / 1000.0;
	if (baseHold == 0.0)
		baseHold = 0.5;
	baselineHoldOff.update(baseHold);

	float64 trigHold = m_dgtzDriver->triggerHoldOff.get(aChannel) / 1000.0;
	if (trigHold == 0.0)
		trigHold = 0.5;
	triggerHoldOff.update(trigHold);

	int32 trigWindow = m_dgtzDriver->triggerWindowSize.get(aChannel);
	if (trigWindow == 0)
		trigWindow = 300;
	triggerWindowSize.update(trigWindow);

	triggerWindowEnable.update(m_dgtzDriver->triggerWindowEnable.get(aChannel));

	digitalGain.update(m_dgtzDriver->digitalGain.get(aChannel));

	int32 trapGain = m_dgtzDriver->trapezoidGain.get(aChannel);
	if (trapGain == 0)
		trapGain = 1;
	trapezoidGain.update(trapGain);

	decimation.update(m_dgtzDriver->decimation.get(aChannel));

	channelActive.update(m_dgtzDriver->channelActive.get(aChannel));

	int32 offSet = m_dgtzDriver->dcOffset.get(aChannel);
	if (offSet == 0)
		offSet = 10;
	dcOffset.update(offSet);

	int32 preTrig = m_dgtzDriver->preTrigger.get(aChannel);
	if (preTrig == 0)
		preTrig = 500;
	preTrigger.update(preTrig);

	pulsePolarity.update(m_dgtzDriver->pulsePolarity.get(aChannel));

	useFirstDerivative.update(m_dgtzDriver->useFirstDerivative.get(aChannel));

	initChannel(aChannel);
}

void CAENPhaController::refreshSignalDecayTime(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->signalDecayTime.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTrapezoidFlatTopProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->trapezoidFlatTop.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTrapezoidRiseTimeProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->trapezoidRiseTime.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshFlatTopDelayProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->flatTopDelay.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTriggerSmoothing(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerSmoothing.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshSignalRiseTimeProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->signalRiseTime.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTriggerThresholdProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerThreshold.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshBaselineMean(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->baselineMean.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshPeakMean(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->peakMean.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshPeaklineHoldOffProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->peakHoldOff.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshBaselineHoldOffProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->baselineHoldOff.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTriggerHoldOffProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerHoldOff.set(displayChannel(), (int32)(aValue * 1000));
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTriggerWindowSizeProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerWindowSize.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTriggerWindowEnableProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerWindowEnable.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshDigitalGainProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->digitalGain.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshTrapezoidGainProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->trapezoidGain.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshDecimationProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->decimation.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshChannelActiveProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->channelActive.set(displayChannel(), aValue);

	// Update channel mask
	int32 mask = 0;
	for (int32 i=0; i<m_dgtzDriver->nbChannels(); ++i) {
		if (m_dgtzDriver->channelActive.get(i))
				mask += (m_dgtzDriver->channelActive.get(i) << i);
	}
	m_dgtzDriver->channelMask = mask;
}

void CAENPhaController::refreshDcOffsetProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->dcOffset.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshPreTriggerProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->preTrigger.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshPulsePolarityProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->pulsePolarity.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::refreshUseFirstDerivativeProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->useFirstDerivative.set(displayChannel(), aValue);
	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::updateBoardProperties() {
	modelName.update(m_dgtzDriver->modelName());
	channels.update(m_dgtzDriver->channels());
	nbChannels.update(m_dgtzDriver->nbChannels());
	minRange = 0;
	maxRange = m_dgtzDriver->nbChannels() - 1;
	nbBits.update(m_dgtzDriver->nbBits());
	dppAcqMode.update(m_dgtzDriver->dppAcqMode());
}

void CAENPhaController::writeParam() {

	m_dgtzDriver.execute(caen_pha::CAENPhaDriver::WRITE_PARAMETERS_COMMAND, false);
}

void CAENPhaController::initBoard() {

	m_dgtzDriver.execute(driver::INIT_COMMAND, false);
}

void CAENPhaController::initChannel(int32 aChannel) {

	m_dgtzDriver->signalDecayTime.set(aChannel, (int32)(signalDecayTime() * 1000));
	m_dgtzDriver->trapezoidFlatTop.set(aChannel, (int32)(trapezoidFlatTop() * 1000));
	m_dgtzDriver->trapezoidRiseTime.set(aChannel, (int32)(trapezoidRiseTime() * 1000));
	m_dgtzDriver->flatTopDelay.set(aChannel, (int32)(flatTopDelay() * 1000));
	m_dgtzDriver->triggerSmoothing.set(aChannel, triggerSmoothing());
	m_dgtzDriver->signalRiseTime.set(aChannel, (int32)(signalRiseTime() * 1000));
	m_dgtzDriver->triggerThreshold.set(aChannel, triggerThreshold());
	m_dgtzDriver->baselineMean.set(aChannel, baselineMean());
	m_dgtzDriver->peakMean.set(aChannel, peakMean());
	m_dgtzDriver->peakHoldOff.set(aChannel, (int32)(peakHoldOff() * 1000));
	m_dgtzDriver->baselineHoldOff.set(aChannel, (int32)(baselineHoldOff() * 1000));
	m_dgtzDriver->triggerHoldOff.set(aChannel, (int32)(triggerHoldOff() * 1000));
	m_dgtzDriver->triggerWindowSize.set(aChannel, triggerWindowSize());
	m_dgtzDriver->triggerWindowEnable.set(aChannel, triggerWindowEnable());
	m_dgtzDriver->digitalGain.set(aChannel, digitalGain());
	m_dgtzDriver->trapezoidGain.set(aChannel, trapezoidGain());
	m_dgtzDriver->decimation.set(aChannel, decimation());
	m_dgtzDriver->channelActive.set(aChannel, channelActive());

	// Update channel mask
	int32 mask = 0;
	for (int32 i=0; i<m_dgtzDriver->nbChannels(); ++i) {
		if (m_dgtzDriver->channelActive.get(i))
				mask += (m_dgtzDriver->channelActive.get(i) << i);
	}
	m_dgtzDriver->channelMask = mask;

	m_dgtzDriver->dcOffset.set(aChannel, dcOffset());
	m_dgtzDriver->preTrigger.set(aChannel, preTrigger());
	m_dgtzDriver->pulsePolarity.set(aChannel, pulsePolarity());
	m_dgtzDriver->useFirstDerivative.set(aChannel, useFirstDerivative());

	boost::thread td(boost::bind(&CAENPhaController::writeParam, this));
}

void CAENPhaController::clearData() {

	commandProgression = 0;
	m_dgtzDriver.execute(caen_pha::CAENPhaDriver::CLEAR_COMMAND, true);
	commandProgression = 100;
}

}
